package me.trixkz.lifestealcore.managers.clans;

import com.mongodb.client.model.Filters;
import com.mongodb.client.model.UpdateOptions;
import lombok.Getter;
import lombok.Setter;
import me.trixkz.lifestealcore.LifeStealCore;
import me.trixkz.lifestealcore.playerdata.PlayerData;
import me.trixkz.lifestealcore.utils.*;
import org.bson.Document;
import org.bukkit.*;
import org.bukkit.entity.Player;

import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

@Getter
@Setter
public class ClanManager {

    private LifeStealCore main = LifeStealCore.getInstance();

    private HashMap<UUID, Clan> clans = new HashMap<UUID, Clan>();

    private Map<UUID, List<UUID>> clanInvites = new TtlHashMap<>(TimeUnit.SECONDS, 15);

    private Cache<List<Clan>> clanLevelsSortCache = new Cache<>(5000);

    public ClanManager() {

    }

    /**
     * Adds an clan to the server for the player
     *
     * @param player - The player
     * @param clan - The clan
     */
    public void createClan(Player player, Clan clan) {
        this.clans.put(player.getUniqueId(), clan);
    }

    /**
     * Removes an clan from the server for the player
     *
     * @param player - The player
     */
    public void removeClan(Player player) {
        this.clans.remove(player.getUniqueId());
    }

    /**
     * Gets the player's clan
     *
     * @param player - The player
     * @return The player's clan
     */
    public Clan getClan(UUID player) {
        return this.clans.get(player);
    }

    /**
     * Checks whether the player is an owner of their clan
     *
     * @param player - The player
     * @return True or false
     */
    public boolean isOwner(OfflinePlayer player) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        return player.getUniqueId().equals(playerData.getClanOwner());
    }

    /**
     * Checks whether the player is in an clan
     *
     * @param player - The player
     * @return True or false
     */
    public boolean isInAClan(OfflinePlayer player) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        return this.clans.containsKey(playerData.getClanOwner()) || (playerData.getClanOwner() != null && this.getClan(playerData.getClanOwner()).getPlayersAdded().contains(player.getUniqueId()));
    }

    /**
     * Checks whether the player is an clan member
     *
     * @param player - The player
     * @return True or false
     */
    public boolean isAClanMember(OfflinePlayer player) {
        return this.main.getClanManager().isOwner(player) || this.main.getClanManager().isInAClan(player);
    }

    /**
     * Checks whether the player's clan has the maximum amount of clan members in it
     *
     * @param player - The player
     * @return True or false
     */
    public boolean hasMaxClanMembers(UUID player) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player);

        return (this.main.getClanManager().getClan(playerData.getClanOwner()).getPlayersAdded().size() + 1) >= this.main.getClanManager().getClan(player).getMaxClanMembers();
    }

    /**
     * Load the player's clan
     *
     * @param player - The player
     */
    public void loadClan(Player player) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        Document document = this.main.getMongoManager().getClans().find(Filters.eq(Constants.CLAN_OWNER, playerData.getClanOwner().toString())).first();

        if (document == null) {
            return;
        }

        String name = document.getString(Constants.CLAN_NAME);

        int level = document.getInteger(Constants.CLAN_LEVEL);

        String ownerString = document.getString(Constants.CLAN_OWNER);

        List<String> playersAddedString = new ArrayList<String>();

        ((List) this.main.getMongoManager().getGson().fromJson(document.getString(Constants.CLAN_PLAYERS_ADDED), this.main.getMongoManager().getType())).forEach(playerAdded -> playersAddedString.add(String.valueOf(playerAdded)));

        boolean locked = document.getBoolean(Constants.CLAN_LOCKED);

        int maxIslandMembers = document.getInteger(Constants.CLAN_MAX_CLAN_MEMBERS);

        UUID owner = UUID.fromString(ownerString);

        List<UUID> playersAdded = new ArrayList<UUID>();

        for (String string : playersAddedString) {
            playersAdded.add(UUID.fromString(string));
        }

        Clan clan = new Clan(
                name,
                level,
                owner,
                playersAdded,
                locked,
                maxIslandMembers
        );

        this.clans.put(owner, clan);

        clan.setLocked(false);
    }

    /**
     * Save the player's clan
     *
     * @param player - The player
     */
    public void saveClan(Player player) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        if (playerData.getClanOwner() == null) {
            return;
        }

        Document document = this.main.getMongoManager().getClans().find(Filters.eq(Constants.CLAN_OWNER, playerData.getClanOwner().toString())).first();

        if (document == null) {
            return;
        }

        Clan clan = this.getClan(playerData.getClanOwner());

        String name = clan.getName();

        int level = clan.getLevel();

        String ownerString = clan.getOwner().toString();

        List<UUID> playersAdded = clan.getPlayersAdded();

        boolean locked = clan.isLocked();

        int maxClanMembers = clan.getMaxClanMembers();

        List<String> playersAddedString = new ArrayList<String>();

        for (UUID uuid : playersAdded) {
            playersAddedString.add(uuid.toString());
        }

        document.put(Constants.CLAN_NAME, name);
        document.put(Constants.CLAN_LEVEL, level);
        document.put(Constants.CLAN_OWNER, ownerString);
        document.put(Constants.CLAN_PLAYERS_ADDED, this.main.getMongoManager().getGson().toJson(playersAddedString));
        document.put(Constants.CLAN_LOCKED, locked);
        document.put(Constants.CLAN_MAX_CLAN_MEMBERS, maxClanMembers);

        Player owner = this.main.getServer().getPlayer(clan.getOwner());

        if (owner != null) {
            owner.chat("/spawn");
        }

        for (UUID uuid : this.main.getClanManager().getClan(playerData.getClanOwner()).getPlayersAdded()) {
            Player playerAdded = this.main.getServer().getPlayer(uuid);

            if (playerAdded != null) {
                playerAdded.chat("/spawn");
            }
        }

        this.clans.remove(playerData.getClanOwner());
        this.main.getMongoManager().getClans().replaceOne(Filters.eq(Constants.CLAN_OWNER, playerData.getClanOwner().toString()), document, new UpdateOptions().upsert(true));

        clan.setLocked(true);
    }

    /**
     * Make the player's clan
     *
     * @param player - The player
     */
    public void makeClan(Player player) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        Document document = new Document();

        if (this.isAClanMember(player)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.CLAN_ALREADY_MADE_MESSAGE));

            return;
        }

        player.sendMessage(Utils.translate(Constants.PREFIX + Constants.MAKING_CLAN_MESSAGE));

        String name = player.getName() + "'s island";

        int level = 0;

        List<UUID> playersAdded = new ArrayList<UUID>();

        boolean locked = false;

        int maxClanMembers = 10;

        String worldName = player.getName() + "Clan";

        Clan clan = new Clan(
                name,
                level,
                player.getUniqueId(),
                playersAdded,
                locked,
                maxClanMembers
        );

        document.put(Constants.CLAN_OWNER, player.getUniqueId().toString());

        this.createClan(player, clan);
        this.main.getMongoManager().getClans().insertOne(document);

        playerData.setClanOwner(player.getUniqueId());

        player.sendMessage(Utils.translate(Constants.PREFIX + Constants.CLAN_MADE_MESSAGE));
    }

    /**
     * Delete the player's clan
     *
     * @param player - The player
     */
    public void deleteIsland(Player player) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        Document document = this.main.getMongoManager().getClans().find(Filters.eq(Constants.CLAN_OWNER, playerData.getClanOwner().toString())).first();

        if (document == null) {
            return;
        }

        if (!this.isOwner(player)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.NOT_CLAN_OWNER_MESSAGE));

            return;
        }

        Clan clan = this.getClan(player.getUniqueId());

        playerData.setClanOwner(null);

        for (UUID uuid : clan.getPlayersAdded()) {
            PlayerData playerAddedPlayerData = this.main.getPlayerDataManager().getOrCreate(uuid);
            playerAddedPlayerData.setClanOwner(null);

            if (this.main.getServer().getPlayer(uuid) == null) {
                this.main.getPlayerDataManager().savePlayerData(playerAddedPlayerData);
                this.main.getPlayerDataManager().getPlayers().remove(uuid);
            }
        }

        clan.sendMessageToAllClanMembers(Utils.translate(Constants.CLAN_PREFIX + Constants.CLAN_DELETED_MESSAGE.replace("%player%", player.getName())));

        this.removeClan(player);
        this.main.getMongoManager().getClans().deleteOne(document);
    }

    /**
     * Transfer the player's clan ownership to the target
     *
     * @param player - The player
     * @param target - The target
     */
    public void transferClanOwnership(Player player, OfflinePlayer target) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        if (!this.isOwner(player)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.NOT_CLAN_OWNER_MESSAGE));

            return;
        }

        if (!this.isAClanMember(target)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.NOT_IN_CLAN_MESSAGE.replace("%player%", target.getName())));

            return;
        }

        Clan clan = this.getClan(player.getUniqueId());
        clan.setOwner(target.getUniqueId());

        playerData.setClanOwner(target.getUniqueId());

        for (UUID uuid : clan.getPlayersAdded()) {
            PlayerData playerAddedPlayerData = this.main.getPlayerDataManager().getPlayerData(uuid);
            playerAddedPlayerData.setClanOwner(target.getUniqueId());

            if (this.main.getServer().getPlayer(uuid) == null) {
                this.main.getPlayerDataManager().savePlayerData(playerAddedPlayerData);
                this.main.getPlayerDataManager().getPlayers().remove(uuid);
            }
        }

        player.sendMessage(Utils.translate(Constants.PREFIX + Constants.CLAN_OWNERSHIP_TRANSFERRED_MESSAGE.replace("%player%", target.getName())));
    }

    /**
     * Look at the player's clan level
     *
     * @param player - The player
     */
    public void lookAtClanLevel(Player player) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        if (!this.isAClanMember(player)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.NO_CLAN_MESSAGE));

            return;
        }

        Clan clan = this.getClan(playerData.getClanOwner());

        player.sendMessage(Utils.translate(Constants.PREFIX + Constants.CLAN_LEVEL_MESSAGE.replace("%level%", String.valueOf(clan.getLevel()))));
    }

    /**
     * Get the clan top
     *
     * @return The list of the clans
     */
    public List<Clan> getClanTop() {
        return this.clanLevelsSortCache.getCache(() -> this.clans.values().stream().sorted(Comparator.comparing(Clan :: getLevel).reversed()).collect(Collectors.toList()));
    }

    public void createClanInvite(Player player, Player target) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        if (!this.isOwner(player)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.NOT_CLAN_OWNER_MESSAGE));

            return;
        }

        if (this.hasClanInvite(player.getUniqueId(), target.getUniqueId())) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.ALREADY_INVITED_TO_CLAN_MESSAGE.replace("%player%", target.getName())));

            return;
        }

        if (this.hasMaxClanMembers(player.getUniqueId())) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.MAXIMUM_AMOUNT_OF_CLAN_MEMBERS_MESSAGE));

            return;
        }

        if (target == null) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.NOT_ONLINE_MESSAGE.replace("%player%", target.getName())));

            return;
        }

        if (this.isAClanMember(target)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.ALREADY_IN_CLAN_MESSAGE.replace("%player%", target.getName())));

            return;
        }

        if (this.getClan(playerData.getClanOwner()).getPlayersAdded().contains(target.getUniqueId())) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.ALREADY_IN_YOUR_CLAN_MESSAGE.replace("%player%", target.getName())));

            return;
        }

        this.clanInvites.computeIfAbsent(target.getUniqueId(), playerTarget -> new ArrayList<>()).add(player.getUniqueId());

        Clickable clickable = new Clickable(Utils.translate(Constants.PREFIX + Constants.INVITED_TO_CLAN_INVITATION_MESSAGE.replace("%player%", player.getName())), Utils.translate(Constants.CLICK_TO_JOIN_CLAN_MESSAGE), "/clan join " + player.getName());
        clickable.sendToPlayer(target);

        player.sendMessage(Utils.translate(Constants.PREFIX + Constants.INVITED_TO_CLAN_MESSAGE.replace("%player%", target.getName())));
    }

    public void removeClanInvite(Player player, OfflinePlayer target) {
        if (!this.isOwner(player)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.NOT_CLAN_OWNER_MESSAGE));

            return;
        }

        if (!this.hasClanInvite(player.getUniqueId(), target.getUniqueId())) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.NOT_INVITED_TO_CLAN_MESSAGE.replace("%player%", target.getName())));

            return;
        }

        this.clanInvites.remove(target.getUniqueId());

        player.sendMessage(Utils.translate(Constants.PREFIX + Constants.UNINVITED_TO_CLAN_MESSAGE.replace("%player%", target.getName())));
    }

    public boolean hasClanInvite(UUID player, UUID target) {
        return this.clanInvites.get(target) != null && this.clanInvites.get(target).contains(player);
    }

    public void joinClan(UUID player, Player target) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player);

        Clan clan = this.getClan(playerData.getClanOwner());

        PlayerData targetPlayerData = this.main.getPlayerDataManager().getPlayerData(target.getUniqueId());

        if (clan == null) {
            target.sendMessage(Utils.translate(Constants.PREFIX + Constants.INVALID_CLAN_MESSAGE));

            return;
        }

        if (this.isAClanMember(target)) {
            target.sendMessage(Utils.translate(Constants.PREFIX + Constants.CLAN_ALREADY_MADE_MESSAGE));

            return;
        }

        if (!this.hasClanInvite(player, target.getUniqueId())) {
            target.sendMessage(Utils.translate(Constants.PREFIX + Constants.NO_INVITE_TO_CLAN_MESSAGE));

            return;
        }

        if (this.hasMaxClanMembers(player)) {
            target.sendMessage(Utils.translate(Constants.PREFIX + Constants.MAXIMUM_AMOUNT_OF_CLAN_MEMBERS_MESSAGE));

            return;
        }

        this.clanInvites.remove(target.getUniqueId());

        clan.getPlayersAdded().add(target.getUniqueId());

        targetPlayerData.setClanOwner(player);

        clan.sendMessageToAllClanMembers(Utils.translate(Constants.CLAN_PREFIX + Constants.JOINED_CLAN_MESSAGE.replace("%player%", target.getName())));

        if (this.main.getServer().getPlayer(player) == null) {
            this.main.getPlayerDataManager().savePlayerData(playerData);
            this.main.getPlayerDataManager().getPlayers().remove(player);
        }
    }

    public void leaveClan(UUID player, Player target) {
        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player);

        if (playerData.getClanOwner() == null) {
            return;
        }

        Clan clan = this.getClan(playerData.getClanOwner());

        if (clan == null) {
            return;
        }

        PlayerData targetPlayerData = this.main.getPlayerDataManager().getPlayerData(target.getUniqueId());

        if (!this.isAClanMember(target)) {
            target.sendMessage(Utils.translate(Constants.PREFIX + Constants.NO_CLAN_MESSAGE));

            return;
        }

        if (this.isOwner(target)) {
            target.sendMessage(Utils.translate(Constants.PREFIX + Constants.INVALID_ACTION_MESSAGE));

            return;
        }

        clan.getPlayersAdded().remove(target.getUniqueId());

        targetPlayerData.setClanOwner(null);

        clan.sendMessageToAllClanMembers(Utils.translate(Constants.CLAN_PREFIX + Constants.LEFT_CLAN_MESSAGE.replace("%player%", target.getName())));

        if (this.main.getServer().getPlayer(player) == null) {
            this.main.getPlayerDataManager().savePlayerData(playerData);
            this.main.getPlayerDataManager().getPlayers().remove(player);
        }
    }

    public void kickPlayerFromClan(Player player, OfflinePlayer target) {
        if (!this.isOwner(player)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.NOT_CLAN_OWNER_MESSAGE));

            return;
        }

        if (this.main.getClanManager().isOwner(target)) {
            player.sendMessage(Utils.translate(Constants.PREFIX + Constants.INVALID_ACTION_MESSAGE));

            return;
        }

        PlayerData playerData = this.main.getPlayerDataManager().getPlayerData(player.getUniqueId());

        Clan clan = this.getClan(playerData.getClanOwner());

        if (this.isAClanMember(target)) {
            PlayerData targetPlayerData = this.main.getPlayerDataManager().getPlayerData(target.getUniqueId());

            clan.getPlayersAdded().remove(target.getUniqueId());

            targetPlayerData.setClanOwner(null);

            clan.sendMessageToAllClanMembers(Utils.translate(Constants.CLAN_PREFIX + Constants.PLAYER_KICKED_FROM_CLAN.replace("%player%", target.getName())));

            return;
        }

        clan.sendMessageToAllClanMembers(Utils.translate(Constants.CLAN_PREFIX + Constants.PLAYER_KICKED_FROM_CLAN.replace("%player%", target.getName())));
    }
}
