package me.trixkz.lifestealcore.managers.koth;

import com.mongodb.BasicDBObject;
import lombok.Getter;
import lombok.Setter;
import me.trixkz.lifestealcore.LifeStealCore;
import me.trixkz.lifestealcore.managers.event.EventScheduledTime;
import me.trixkz.lifestealcore.utils.CustomLocation;
import me.trixkz.lifestealcore.utils.Utils;
import me.trixkz.lifestealcore.utils.config.FileConfig;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.scheduler.BukkitRunnable;
import java.io.File;
import java.util.*;

@Getter
@Setter
public class KOTHManager {

    private LifeStealCore main = LifeStealCore.getInstance();

    private FileConfig kothsConfig = this.main.getKothsConfig();

    private Map<String, KOTH> koths = new HashMap<String, KOTH>();
    private Map<EventScheduledTime, String> kothSchedule = new HashMap<EventScheduledTime, String>();

    private boolean kothScheduledEnabled;

    public KOTHManager() {
        this.loadKOTHs();
        this.loadKOTHSchedules();

        new BukkitRunnable() {
            @Override
            public void run() {
                for (KOTH koth : koths.values()) {
                    koth.tick();
                }
            }
        }.runTaskTimer(this.main, 5L, 20L);

        this.main.getServer().getScheduler().runTaskTimer(this.main, () -> {
            this.enableKOTHs();
        }, 20L, 20L);
    }

    public KOTH getKOTH(String name) {
        return this.koths.get(name);
    }

    public void makeKOTH(String name, KOTH koth) {
        this.koths.put(name, koth);
    }

    public void removeKOTH(String name) {
        this.koths.remove(name);
    }

    public void loadKOTHs() {
        FileConfiguration fileConfiguration = this.kothsConfig.getConfig();

        ConfigurationSection configurationSection = fileConfiguration.getConfigurationSection("koths");

        if (configurationSection == null) {
            return;
        }

        configurationSection.getKeys(false).forEach(name -> {
            String world = configurationSection.getString(name + ".world");

            CustomLocation customLocation = CustomLocation.stringToLocation(configurationSection.getString(name + ".custom-location"));

            int time = configurationSection.getInt(name + ".time");
            int distance = configurationSection.getInt(name + ".distance");

            boolean active = configurationSection.getBoolean(name + ".active");
            boolean hidden = configurationSection.getBoolean(name + ".hidden");

            KOTH koth = new KOTH(name, world, customLocation, time, distance, active, hidden);

            this.koths.put(name, koth);
        });

        this.koths.values().stream().filter(KOTH:: isActive).findFirst().ifPresent((koth) -> {
            koth.setActive(false);

            this.main.getServer().getScheduler().runTaskLater(this.main, () -> {
                if (this.koths.values().stream().noneMatch(KOTH:: isActive)) {
                    koth.enable();
                }
            }, 300L);
        });
    }

    public void saveKOTHs() {
        FileConfiguration fileConfiguration = this.kothsConfig.getConfig();
        fileConfiguration.set("koths", null);

        this.koths.forEach((name, koth) -> {
            String world = koth.getWorld();
            String customLocation = CustomLocation.locationToString(koth.getCustomLocation());

            int time = koth.getTime();
            int distance = koth.getDistance();

            boolean active = koth.isActive();
            boolean hidden = koth.isHidden();

            String root = "koths." + name;

            fileConfiguration.set(root + ".world", world);
            fileConfiguration.set(root + ".custom-location", customLocation);
            fileConfiguration.set(root + ".time", time);
            fileConfiguration.set(root + ".distance", distance);
            fileConfiguration.set(root + ".active", active);
            fileConfiguration.set(root + ".hidden", hidden);
        });

        this.kothsConfig.save();
    }

    public void loadKOTHSchedules() {
        this.kothSchedule.clear();

        try {
            File file = new File(this.main.getDataFolder(), "event-schedule.json");

            if (!file.exists()) {
                file.createNewFile();

                BasicDBObject basicDBObjectOne = new BasicDBObject();

                int currentDay = Calendar.getInstance().get(Calendar.DAY_OF_YEAR);

                List<String> koths = new ArrayList<String>();

                for (KOTH koth : this.koths.values()) {
                    if (koth.isHidden()) {
                        continue;
                    }

                    koths.add(koth.getName());
                }

                for (int dayOffset = 0; dayOffset < 21; dayOffset++) {
                    int day = (currentDay + dayOffset) % 365;

                    EventScheduledTime[] eventScheduledTimes = new EventScheduledTime[]{
                            new EventScheduledTime(day, 0, 0), // 00:30am EST
                            new EventScheduledTime(day, 1, 0), // 02:00am EST
                            new EventScheduledTime(day, 2, 0), // 03:30am EST
                            new EventScheduledTime(day, 3, 0), // 05:00am EST
                            new EventScheduledTime(day, 4, 0), // 06:30am EST
                            new EventScheduledTime(day, 5, 0), // 08:00am EST
                            new EventScheduledTime(day, 6, 0), // 09:30am EST
                            new EventScheduledTime(day, 7, 0), // 11:30am EST
                            new EventScheduledTime(day, 8, 0), // 01:30pm EST
                            new EventScheduledTime(day, 9, 0), // 15:30pm EST
                            new EventScheduledTime(day, 10, 0), // 17:00pm EST
                            new EventScheduledTime(day, 11, 0), // 18:30pm EST
                            new EventScheduledTime(day, 12, 0), // 20:00am EST
                            new EventScheduledTime(day, 13, 0), // 21:30pm EST
                            new EventScheduledTime(day, 14, 0), // 21:30pm EST
                            new EventScheduledTime(day, 15, 0), // 21:30pm EST
                            new EventScheduledTime(day, 16, 0), // 21:30pm EST
                            new EventScheduledTime(day, 17, 0), // 21:30pm EST
                            new EventScheduledTime(day, 18, 0), // 21:30pm EST
                            new EventScheduledTime(day, 19, 0), // 21:30pm EST
                            new EventScheduledTime(day, 20, 0), // 21:30pm EST
                            new EventScheduledTime(day, 21, 0), // 21:30pm EST
                            new EventScheduledTime(day, 22, 0), // 21:30pm EST
                            new EventScheduledTime(day, 23, 0) // 21:30pm EST
                    };

                    Collections.shuffle(koths);

                    if (!koths.isEmpty()) {
                        for (int i = 0; i < eventScheduledTimes.length; i++) {
                            EventScheduledTime eventScheduledTime = eventScheduledTimes[i];

                            String name = koths.get(i % koths.size());

                            basicDBObjectOne.put(eventScheduledTime.toString(), name);
                        }
                    }
                }

//                FileUtils.write(file, JSON.serialize(basicDBObjectOne.toString()));
            }

//            BasicDBObject basicDBObjectTwo = (BasicDBObject) JSON.parse(FileUtils.readFileToString(file));
//
//            if (basicDBObjectTwo == null) {
//                return;
//            }
//
//            for (Map.Entry<String, Object> mapEntry : basicDBObjectTwo.entrySet()) {
//                EventScheduledTime eventScheduledTime = EventScheduledTime.parse(mapEntry.getKey());
//
//                this.kothSchedule.put(eventScheduledTime, String.valueOf(mapEntry.getValue()));
//            }
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    public void enableKOTHs() {
        for (KOTH koth : this.koths.values()) {
            if (koth.isActive()) {
                return;
            }
        }

        EventScheduledTime eventScheduledTime = EventScheduledTime.parse(new Date());

        if (this.kothSchedule.containsKey(eventScheduledTime)) {
            String name = this.kothSchedule.get(eventScheduledTime);

            KOTH koth = this.koths.get(name);

            if (this.main.getServer().getOnlinePlayers().size() < 1) {
                this.kothSchedule.remove(eventScheduledTime);
                this.main.getServer().broadcastMessage(Utils.translate("&cA KOTH would of started, but there are less than 1 player online"));

                return;
            }

            koth.enable();
        }
    }

    public void setEventSchedule() {
        List<String> koths = new ArrayList<String>();

        for (KOTH koth : this.koths.values()) {
            if (koth.isHidden()) {
                continue;
            }

            koths.add(koth.getName());
        }

        for (int minute = 0; minute < 60; minute++) {
            for (int hour = 0; hour < 24; hour++) {
                this.kothSchedule.put(new EventScheduledTime(Calendar.getInstance().get(Calendar.DAY_OF_YEAR), (Calendar.getInstance().get(Calendar.HOUR_OF_DAY) + hour) % 24, minute), koths.get(0));
            }
        }
    }
}
